from cas_metadata_tools.base_constants import BaseConstants

class EXIFConstants(BaseConstants):
    # EXIF Constants
    EXIF_ARTIST = "EXIF:Artist"
    EXIF_COPYRIGHT = "EXIF:Copyright"
    EXIF_CREATE_DATE = "EXIF:CreateDate"
    EXIF_IFD0_IMAGE_DESCRIPTION = "EXIF:IFD0:ImageDescription"
    EXIF_IMAGE_DESCRIPTION = "EXIF:ImageDescription"

    # IDF0 constants
    IFD0_COPYRIGHT = "IFD0:Copyright"

    # IPTC Constants
    IPTC_BY_LINE = "IPTC:By-line"
    IPTC_BY_LINE_TITLE = "IPTC:By-lineTitle"
    IPTC_CAPTION_ABSTRACT = "IPTC:Caption-Abstract"
    IPTC_COPYRIGHT_NOTICE = "IPTC:CopyrightNotice"
    IPTC_CREDIT = "IPTC:Credit"
    IPTC_KEYWORDS = "IPTC:Keywords"

    # Photoshop Constants
    PHOTOSHOP_COPYRIGHT_FLAG = "Photoshop:CopyrightFlag"
    PHOTOSHOP_CREDIT = "XMP-photoshop:Credit"

    # XMP Constants
    XMP_CREATE_DATE = "XMP:CreateDate"
    XMP_CREATOR = "XMP:Creator"
    XMP_CREATOR_ADDRESS = "XMP:CreatorAddress"
    XMP_CREATOR_CITY = "XMP:CreatorCity"
    XMP_CREATOR_COUNTRY = "XMP:CreatorCountry"
    XMP_CREATOR_POSTAL_CODE = "XMP:CreatorPostalCode"
    XMP_CREATOR_REGION = "XMP:CreatorRegion"
    XMP_CREATOR_WORK_URL = "XMP:CreatorWorkURL"
    XMP_CREDIT = "XMP:Credit"
    XMP_DATE_CREATED = "XMP:DateCreated"
    XMP_DC_DESCRIPTION = "XMP-dc:Description"
    XMP_DC_SUBJECT = "XMP-dc:Subject"
    XMP_DESCRIPTION = "XMP:Description"
    XMP_LABEL = "XMP:Label"
    XMP_LR_HIERARCHICAL_SUBJECT = "XMP-lr:HierarchicalSubject"
    XMP_PLUS_IMAGE_SUPPLIER_NAME = "XMP-plus:ImageSupplierName"
    XMP_RIGHTS = "XMP:Rights" # this is the same as XMP-dc:Rights
    XMP_RIGHTS_USAGE_TERMS = "XMP-xmpRights:UsageTerms"
    XMP_TITLE = "XMP:Title"
    XMP_USAGE = "XMP:Usage"
    XMP_USAGE_TERMS = "XMP:UsageTerms"



class BaseConstants:
    @classmethod
    def iterate_constants(cls):
        """
        Iterates over all constants.
        """
        for key, value in cls.__dict__.items():
            if not key.startswith("__") and not callable(value):
                yield key, value

    @classmethod
    def check_dict_valid(cls, dict: dict):
        valid_dict = {}
        for key in dict.keys():
            value = cls.is_valid_constant(key)
            valid_dict[key] = value
        return valid_dict


    @classmethod
    def is_valid_constant(cls, value):
        """
        Checks if the provided value is valid within the constants.
        """
        return value in cls.__dict__.values()


    @classmethod
    def get_constant(cls, key):
        """
        Returns the value of the constant key.
        """
        return cls.__dict__.get(key)

    @classmethod
    def query_by_substring(cls, substring):
        """
        Queries the constants for keys/values containing the provided substring.
        """
        result = {}
        for key, value in cls.__dict__.items():
            if not key.startswith("__") and not callable(value):
                if substring.lower() in key.lower() or substring.lower() in value.lower():
                    result[key] = value
        return result



"""metadata_tools: utility functions for the addition, removal and reading of iptc and exif image metadata"""
from wrapt_timeout_decorator import timeout
import errno
import os
import logging
import subprocess
import traceback
from cas_metadata_tools.EXIF_constants import EXIFConstants

class MetadataTools:

    def __init__(self, path):
        self.path = path
        self.logger = logging.getLogger('MetadataTools')
        self.env = os.environ.copy()
    @staticmethod
    def safe_decode(byte_data):
        try:
            return byte_data.decode('utf-8')
        except UnicodeDecodeError:
            try:
                return byte_data.decode('ISO-8859-1')
            except UnicodeDecodeError:
                return byte_data.decode('latin-1')
            finally:
                return byte_data.decode('utf-8', errors='ignore')
        

    @timeout(20, os.strerror(errno.ETIMEDOUT))
    def read_exif_tags(self):

        """Reads all EXIF tags from an image using ExifTool with advanced formatting and returns them as a dictionary."""
        command = ['exiftool', '-a', '-g', '-G', self.path]
        try:
            result = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE,text=False, env=self.env)
            result.stdout = MetadataTools.safe_decode(result.stdout)
            result.stderr = MetadataTools.safe_decode(result.stderr)
            if result.stderr:
                raise ValueError(f"ExifTool error: {result.stderr.strip()}")
            tags = {}
            for line in result.stdout.split("\n"):
                if ": " in line:
                    group, key_value = line.split("]", 1)
                    key, value = key_value.split(":", 1)
                    formatted_group = group.replace('[', '').strip()
                    formatted_key = key.replace(' ', '').strip()
                    if value.strip():
                        tags[formatted_group + ':' + formatted_key] = value.strip()
            return tags
        except Exception as e:
            traceback.print_exc()
            raise ValueError(f"Command returned with error: {e}")
        finally:
            self.logger.info("EXIF data read successfully")

    @timeout(20, os.strerror(errno.ETIMEDOUT))
    def write_exif_tags(self, exif_dict, overwrite_blank=False):
        """Writes all exif tags to an image with a single call to ExifTool"""
        self.logger.info(f"Processing EXIF data for: {self.path}")

        valid_dict = EXIFConstants.check_dict_valid(dict=exif_dict)

        if any(value is False for value in valid_dict.values()):
            raise ValueError(f"Invalid keys in exif_dict, check exif constants:{valid_dict}")

        args = ["exiftool", "-overwrite_original"]
        if overwrite_blank:
            valid_args = [f"-{key}={value}" if value is not None else f"-{key}= " for key, value in exif_dict.items()]
        else:
            valid_args = [f"-{key}={value}" for key, value in exif_dict.items() if value is not None]

        if not valid_args:
            return

        args.extend(valid_args)
        args.append(self.path)
        try:
            subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE,env=self.env)
        except Exception as e:
            traceback.print_exc()
            raise ValueError(f"ExifTool command returned with error: {e}")
        self.logger.info("EXIF data added successfully")




import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import unittest
import pandas as pd
import shutil
from datetime import datetime
from cas_metadata_tools import MetadataTools
import platform



class TestMetadataTools(unittest.TestCase):

    def setUp(self):
        self.path = "tests/test_images/test_image.jpg"
        if platform.system() == "Darwin":
            encoding = "en_US.UTF-8"
        else:
            encoding = "C.UTF-8"

        self.md = MetadataTools(path=self.path)

        shutil.copyfile("tests/test_images/test_image.jpg", "tests/test_images/image_backup.jpg")


    def test_read_exif_tags(self):
        """tests exif read function"""
        exif_dict = self.md.read_exif_tags()
        self.assertFalse(pd.isna(exif_dict))
        self.assertNotEqual(exif_dict, {})
        self.assertEqual(exif_dict['EXIF:LensMake'], 'Apple')

    def test_write_exif_tags(self):
        """tests exif attach function"""
        datetime_test = datetime.now()
        datetime_test = datetime_test.strftime('%Y:%m:%d %H:%M:%S.%f')
        exif_dict = {"XMP:CreatorCity": 'San Francisco', 'IPTC:CopyrightNotice': '\u00A9 \u00B0 CaliforniaAcademy',
                     'XMP:CreateDate': f'{datetime_test}'}
        self.md.write_exif_tags(exif_dict=exif_dict)
        exif_return = self.md.read_exif_tags()

        self.assertEqual("San Francisco", exif_return['XMP:CreatorCity'])
        self.assertEqual('\u00A9 \u00B0 CaliforniaAcademy', exif_return['IPTC:CopyrightNotice'])
        self.assertEqual(str(f'{datetime_test}'), str(exif_return['XMP:CreateDate']))

    def test_invalid_exif_tags(self):
        exif_dict = {'EXIF:TEST1': 'Samsung', 'IPTC:CopyrightNotice': '\u00A9 \u00B0 CaliforniaAcademy', 'EXIF:ApertureValid': '1.5'}
        with self.assertRaises(ValueError) as context:
            self.md.write_exif_tags(exif_dict=exif_dict)

        self.assertEqual(str(context.exception),
                         "Invalid keys in exif_dict, check exif "
                         "constants:{'EXIF:TEST1': False, 'IPTC:CopyrightNotice': True, 'EXIF:ApertureValid': False}")

    def test_latin_1_encoding(self):
        self.md.path = os.path.join("tests", "test_images", "test_image_latin1.jpg")
        exif_dict = self.md.read_exif_tags()
        self.assertEqual(exif_dict['EXIF:Artist'], 'Johanna Loacker')
        # this is latin1 encoded
        self.assertEqual(exif_dict['EXIF:Copyright'], '© California Academy of Sciences licensed under CC BY-NC-SA')
    
    def test_other_encoding(self):
        self.md.path = os.path.join("tests", "test_images", "exif_gb2312.jpg")
        exif_dict = self.md.read_exif_tags()
        self.assertEqual(exif_dict['EXIF:UserComment'], '你好！')

    def tearDown(self):
        del self.md
        shutil.copyfile("tests/test_images/image_backup.jpg", "tests/test_images/test_image.jpg")


if __name__ == '__main__':
    unittest.main()
